# SPDX-License-Identifier: EUPL-1.2
import argparse
import logging
from pathlib import Path
from typing import List
from lazydog_merge_sqlite import merge_sqlite

logger = logging.getLogger("lazydog_merge_main")


def check_db_files(db_files: List[Path]):
    """Check if a list of paths point to existing files on disk"""
    for dbfile in db_files:
        if not dbfile.exists():
            logger.error(f"File {dbfile} does not exist and can't be loaded")
            exit(1)
        if not dbfile.is_file():
            logger.error(f"File {dbfile} is not a file can't be loaded")
            exit(1)


def main():
    parser = argparse.ArgumentParser(
        prog="lazydog_merge_sqlite",
        description="Merge Sqlite DB files even when they have different tables",
    )
    parser.add_argument(
        "db_files",
        metavar="N",
        nargs="+",
        help="list of existing DB files to merge",
        type=Path,
    )
    parser.add_argument("--out", help="Output DB to write to", type=Path)
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Write more output"
    )
    args = parser.parse_args()
    if args.verbose:
        log_level = "DEBUG"
    else:
        log_level = "INFO"
    logging.basicConfig(
        format="%(asctime)s %(name)s %(levelname)s: %(message)s", level=log_level
    )
    check_db_files(args.db_files)

    if args.out is None:
        if len(args.db_files) == 1:
            logger.warn("Can't merge anything for a single file")
            exit(0)
        merge_sqlite.sql_merge_dbs(args.db_files[1:], args.db_files[0])

    else:
        dst: Path = args.out
        if dst.exists():
            if not dst.is_file():
                logger.error("Output location already exists and isn't a file")
                exit(1)
            merge_sqlite.sql_merge_dbs(args.db_files, dst)
        else:
            merge_sqlite.first_copy(args.db_files[0], dst)
            merge_sqlite.sql_merge_dbs(args.db_files[1:], dst)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: EUPL-1.2
from pathlib import Path
from lazydog_merge_sqlite import merge_sqlite
import sqlite3


def check_tool_info_merger(db_path):
    "Get all tools found"
    conn = sqlite3.connect(db_path)
    curs = conn.cursor()
    res = curs.execute("SELECT tool_name FROM tool_info")
    return set([x[0] for x in res.fetchall()])


def test_table_info():
    a_db_path = (Path(__file__).parent) / "data" / "openScale.db"
    all_info = merge_sqlite.get_table_info(a_db_path)
    assert len(all_info) == 2
    assert len(all_info["openscale_measurements"]) == 4


def test_table_comparison():
    columns_a = [
        merge_sqlite.Column(0, "a", "REAL", True, 1),
        merge_sqlite.Column(1, "b", "TEXT", False, 0),
        merge_sqlite.Column(2, "c", "REAL", False, 0),
    ]
    a = set(columns_a)
    columns_b = [
        merge_sqlite.Column(1, "b", "TEXT", False, 0),
        merge_sqlite.Column(0, "a", "REAL", True, 1),
        merge_sqlite.Column(2, "c", "REAL", False, 0),
    ]
    b = set(columns_b)
    assert a == b
    db_1 = {"tab_a": columns_a}
    db_2 = {"tab_a": columns_b}
    check1 = merge_sqlite.check_compatibility(db_1, db_2)
    assert check1.mergable == True
    assert check1.no_overlap == False
    assert check1.overlapping_tables == ["tab_a"]
    assert check1.non_overlapping_tables == []

    columns_c = [
        merge_sqlite.Column(1, "b", "TEXT", False, 0),
        merge_sqlite.Column(0, "a", "REAL", True, 1),
        merge_sqlite.Column(2, "c", "REAL", True, 0),
    ]
    db_1 = {"tab_a": columns_a}
    db_2 = {"tab_a": columns_c}
    check2 = merge_sqlite.check_compatibility(db_1, db_2)
    assert check2.mergable == False
    assert check2.no_overlap == False
    assert check2.overlapping_tables == []  # Nothing, bad match
    assert check2.non_overlapping_tables == []

    db_1 = {"tab_a": columns_a}
    db_2 = {"tab_b": columns_c}
    check3 = merge_sqlite.check_compatibility(db_1, db_2)
    assert check3.mergable == True
    assert check3.no_overlap == True
    assert check3.overlapping_tables == []
    assert check3.non_overlapping_tables == ["tab_a"]


def test_first_copy(db_export_file):
    a_db_path = (Path(__file__).parent) / "data" / "openScale.db"

    merge_sqlite.first_copy(a_db_path, db_export_file)

    table_info = merge_sqlite.get_table_info(db_export_file)
    assert "openscale_measurements" in table_info


def test_merger(db_export_file):
    a_db_path = (Path(__file__).parent) / "data" / "healthkit.db"
    merge_sqlite.first_copy(a_db_path, db_export_file)

    b_db_path = (Path(__file__).parent) / "data" / "paseo.db"

    a_info = merge_sqlite.get_table_info(db_export_file)
    b_info = merge_sqlite.get_table_info(b_db_path)
    compat_res = merge_sqlite.check_compatibility(b_info, a_info)
    assert compat_res.mergable == True
    assert compat_res.no_overlap == False
    merge_sqlite.merge_dbs(
        b_db_path,
        db_export_file,
        compat_res.overlapping_tables,
        compat_res.non_overlapping_tables,
    )

    out_info = merge_sqlite.get_table_info(db_export_file)
    assert "healthkit_Height" in out_info
    assert "paseo_hours" in out_info

    tools_found = check_tool_info_merger(db_export_file)
    assert "paseo-to-lazydog" in tools_found
    assert "healthkit-to-sqlite" in tools_found


# SPDX-License-Identifier: EUPL-1.2
import sqlite3
import logging
from typing import List, Dict
from dataclasses import dataclass
from pathlib import Path

COMMON_SQLITE_SUFFIX = {".sqlite", ".db"}
logger = logging.getLogger("sqlite_merger")


@dataclass(frozen=True)
class Column:
    """Column in a Sqlite table"""

    cid: int
    name: str
    type_: str
    notnull: bool
    pk: int


TABLE_NAME_QUERY = """
SELECT 
    name
FROM 
    sqlite_schema
WHERE 
    type ='table' AND 
    name NOT LIKE 'sqlite_%';
"""


def run_table_info_query(curs: sqlite3.Cursor, table_name: str) -> List[Column]:
    """Get Table information for a particular table"""

    def map_tup(t) -> Column:
        return Column(t[0], t[1], t[2], bool(t[3]), t[5])

    res = curs.execute("pragma table_info(" + table_name + ")")
    return [map_tup(t) for t in res.fetchall()]


def get_table_info(db_loc: Path) -> Dict[str, List[Column]]:
    """Retrieve the list of tables can be moved"""
    if db_loc.suffix not in COMMON_SQLITE_SUFFIX:
        logger.warn("Database filetype might be wrong: " + db_loc.suffix)
    conn = sqlite3.connect(db_loc)
    curs = conn.cursor()
    res = curs.execute(TABLE_NAME_QUERY)
    tables = res.fetchall()
    out_tables = {}
    for (table_name,) in tables:
        table_info = run_table_info_query(curs, table_name)
        out_tables[table_name] = table_info
    conn.close()
    return out_tables


@dataclass(frozen=True)
class CompatResult:
    no_overlap: bool
    mergable: bool
    overlapping_tables: List[str]
    non_overlapping_tables: List[str]
    failed_table: str = ""


def check_compatibility(
    src_db_info: Dict[str, List[Column]], dst_db_info: Dict[str, List[Column]]
) -> CompatResult:
    """Check if two databases are compatible and can be merged
    Returns results from the tables of a's perspective"""
    src_keys = set(src_db_info.keys())
    dst_keys = set(dst_db_info.keys())
    matching_tables = src_keys.intersection(dst_keys)
    if len(matching_tables) == 0:
        return CompatResult(True, True, [], list(src_keys))
    for tab in matching_tables:
        a_cols = set(src_db_info[tab])
        b_cols = set(dst_db_info[tab])
        if a_cols != b_cols:
            return CompatResult(False, False, [], [], tab)
    non_matching_tables = src_keys.difference(matching_tables)
    assert matching_tables.union(non_matching_tables) == src_keys
    return CompatResult(False, True, list(matching_tables), list(non_matching_tables))


def first_copy(src: Path, dst: Path):
    """Quickly copy the complete src db into a fresh dst db"""
    src_conn = sqlite3.connect(src)
    dst_conn = sqlite3.connect(dst)
    logger.debug(f"Running backup: {src} → {dst}")
    src_conn.backup(dst_conn, pages=4)
    dst_conn.close()
    src_conn.close()


def merge_dbs(
    src: Path,
    dst: Path,
    overlapping_src_tables: List[str],
    non_overlap_src_tables: List[str],
):
    """Copy all tables from the src db into a dst db"""
    dst_conn = sqlite3.connect(dst)

    dst_cur = dst_conn.cursor()
    dst_cur.execute(f"ATTACH DATABASE '{str(src.absolute())}' AS dba")

    for non_ovr_st in non_overlap_src_tables:
        q = f"CREATE TABLE {non_ovr_st} AS SELECT * FROM dba.{non_ovr_st}"
        logger.debug("Running non overlap: " + q)
        dst_cur.execute(q)
        dst_conn.commit()

    for ovr_st in overlapping_src_tables:
        q = f"""
        INSERT INTO {ovr_st}
        SELECT *
        FROM dba.{ovr_st}
        """
        logger.debug("Running overlap: " + q)
        try:
            dst_cur.execute(q)
            dst_conn.commit()
        except sqlite3.IntegrityError as e:
            logger.error(e)

    dst_conn.close()


def sql_merge_dbs(rest: List[Path], dst: Path):
    """Merge DBs with data. copies tables from rest to dst"""
    for other_db in rest:
        dst_info = get_table_info(dst)
        other_info = get_table_info(other_db)
        compat = check_compatibility(other_info, dst_info)
        if not compat.mergable:
            logger.error(
                f"Can't merge {other_db} into {dst}\n"
                + f"The table {compat.failed_table} can't be matched"
            )
            continue
        merge_dbs(
            other_db, dst, compat.overlapping_tables, compat.non_overlapping_tables
        )
